// The MIT License (MIT)
//
// Copyright (c) 2015 Koto Furumiya
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

// NOTE: DO NOT do your own optimization without any benchmark.

/// Utilities for handling Japanese texts.
library japanese_util;

import 'dart:collection';

const Map<String, String> _hiragana2Katakana =
    const <String, String>{"あ": "ア", "い": "イ", "う": "ウ", "え": "エ", "お": "オ",
                           "か": "カ", "き": "キ", "く": "ク", "け": "ケ", "こ": "コ",
                           "さ": "サ", "し": "シ", "す": "ス", "せ": "セ", "そ": "ソ",
                           "た": "タ", "ち": "チ", "つ": "ツ", "て": "テ", "と": "ト",
                           "な": "ナ", "に": "ニ", "ぬ": "ヌ", "ね": "ネ", "の": "ノ",
                           "は": "ハ", "ひ": "ヒ", "ふ": "フ", "へ": "ヘ", "ほ": "ホ",
                           "ま": "マ", "み": "ミ", "む": "ム", "め": "メ", "も": "モ",
                           "や": "ヤ",  "ゆ": "ユ", "よ": "ヨ",
                           "ら": "ラ", "り": "リ", "る": "ル", "れ": "レ", "ろ": "ロ",
                           "わ": "ワ", "ゐ": "ヰ", "ゑ": "ヱ", "を": "ヲ",
                           "ん": "ン",
                           "が": "ガ", "ぎ": "ギ", "ぐ": "グ", "げ": "ゲ", "ご": "ゴ",
                           "ざ": "ザ", "じ": "ジ", "ず": "ズ", "ぜ": "ゼ", "ぞ": "ゾ",
                           "だ": "ダ", "ぢ": "ヂ", "づ": "ヅ", "で": "デ", "ど": "ド",
                           "ゔ": "ヴ",
                           "ぱ": "パ", "ぴ": "ピ", "ぷ": "プ", "ぺ": "ペ", "ぽ": "ポ",
                           "ぁ": "ァ", "ぃ": "ィ", "ぅ": "ゥ", "ぇ": "ェ", "ぉ": "ォ",
                           "っ": "ッ",
                           "ゃ": "ャ", "ゅ": "ュ", "ょ": "ョ",
                           "ゎ": "ヮ"};

const Map<String, String> _katakana2Hiragana =
    const <String, String>{"ア": "あ", "イ": "い", "ウ": "う", "エ": "え", "オ": "お",
                           "カ": "か", "キ": "き", "ク": "く", "ケ": "け", "コ": "こ",
                           "サ": "さ", "シ": "し", "ス": "す", "セ": "せ", "ソ": "そ",
                           "タ": "た", "チ": "ち", "ツ": "つ", "テ": "て", "ト": "と",
                           "ナ": "な", "ニ": "に", "ヌ": "ぬ", "ネ": "ね", "ノ": "の",
                           "ハ": "は", "ヒ": "ひ", "フ": "ふ", "ヘ": "へ", "ホ": "ほ",
                           "マ": "ま", "ミ": "み", "ム": "む", "メ": "め", "モ": "も",
                           "ヤ": "や", "ユ": "ゆ", "ヨ": "よ",
                           "ラ": "ら", "リ": "り", "ル": "る", "レ": "れ", "ロ": "ろ",
                           "ワ": "わ", "ヰ": "ゐ", "ヱ": "ゑ", "ヲ": "を",
                           "ン": "ん",
                           "ガ": "が", "ギ": "ぎ", "グ": "ぐ", "ゲ": "げ", "ゴ": "ご",
                           "ザ": "ざ", "ジ": "じ", "ズ": "ず", "ゼ": "ぜ", "ゾ": "ぞ",
                           "ダ": "だ", "ヂ": "ぢ", "ヅ": "づ", "デ": "で", "ド": "ど",
                           "ヴ": "ゔ",
                           "パ": "ぱ", "ピ": "ぴ", "プ": "ぷ", "ペ": "ぺ", "ポ": "ぽ",
                           "ァ": "ぁ", "ィ": "ぃ", "ゥ": "ぅ", "ェ": "ぇ", "ォ": "ぉ",
                           "ッ": "っ",
                           "ャ": "ゃ", "ュ": "ゅ", "ョ": "ょ",
                           "ヮ": "ゎ"};

/// Converts katakanas in the [text] into hiragana.
///
///     hiraganize("いろは歌イロハ歌abc"); // "いろは歌いろは歌abc"
String hiraganize(String text) => new JapaneseString(text).hiraganize().string;

/// Converts hiraganas in the [text] into katakana.
///
///     katakanize("いろは歌イロハ歌abc"); // "イロハ歌イロハ歌abc"
String katakanize(String text) => new JapaneseString(text).katakanize().string;

/// A class for getting characters, one at a time, from a String.
///
/// The iterator considers a surrogate pair as a single character.
///
/// A typical usage:
///
///     var iterator = new JapaneseIterator("いろはにほへと");
///     while(iterator.moveNext()) {
///       print(iterator.current);
///     }
class JapaneseIterator implements Iterator<String> {
  final RuneIterator _runeIterator;

  JapaneseIterator(String text) : _runeIterator = text.runes.iterator;

  bool moveNext() => _runeIterator.moveNext();

  String get current => new String.fromCharCode(_runeIterator.current);
}

/// A sequence of Japanese characters.
///
/// The JapaneseString regards a surrogate pair as a character.
/// Therefore, you can easily handle each Japanese characters separately.
///
/// For example, you can count the number of Japanese characters correctly:
///
///     var str = "𩸽のお寿司🍣";
///     var jStr = new JapaneseString(str);
///     print(str.length);  // outputs 8(incorrect).
///     print(jStr.length); // outputs 6(correct).
class JapaneseString extends IterableBase<String> {
  /// The raw string.
  final String string;
  final JapaneseIterator iterator;

  /// Creates a new JapaneseString instance with the [string].
  JapaneseString(String string) :
      this.string = string,
      iterator = new JapaneseIterator(string);

  JapaneseString _replaceTextWith(Map<String, String> map) {
    var buffer = new StringBuffer();

    this.forEach((char) {
      var replaced = map.containsKey(char) ? map[char] : char;
      buffer.write(replaced);
    });

    return new JapaneseString(buffer.toString());
  }

  /// Converts katakanas in the string into hiragana.
  ///
  ///     var str = new JapaneseString("いろは歌イロハ歌abc");
  ///     print(str.hiraganize()); // "いろは歌いろは歌abc"
  JapaneseString hiraganize() => _replaceTextWith(_katakana2Hiragana);

  /// Converts hiraganas in the string into katakana.
  ///
  ///     var str = new JapaneseString("いろは歌イロハ歌abc");
  ///     print(str.katakanize()); // "イロハ歌イロハ歌abc"
  JapaneseString katakanize() => _replaceTextWith(_hiragana2Katakana);

  @override
  bool operator==(other) {
    if(other is String) return string == other;
    if(other is JapaneseString) return string == other.string;
    return false;
  }

  @override
  int get hashCode => string.hashCode;
}